/* GEM-View 3.00 PRINTER Modul: Idee fr Farbdrucker mit 8-Farb-Dither
   (c) 22.10.1993 by Dieter Fiebelkorn
                     Grner Weg 29a
                     D-45768 Marl
                     (GERMANY)
*/

#include <stdio.h>
#include "image.h"
#include "moduls.h"

typedef size_t    bit32;
int     errno;

#define rtable(name,n)\
	static bit32 name[16] =\
	{\
		0x00000000UL<<n, 0x01000000UL<<n, 0x00010000UL<<n, 0x01010000UL<<n,\
		0x00000100UL<<n, 0x01000100UL<<n, 0x00010100UL<<n, 0x01010100UL<<n,\
		0x00000001UL<<n, 0x01000001UL<<n, 0x00010001UL<<n, 0x01010001UL<<n,\
		0x00000101UL<<n, 0x01000101UL<<n, 0x00010101UL<<n, 0x01010101UL<<n,\
	};

rtable(rtab0,0)
rtable(rtab1,1)
rtable(rtab2,2)
rtable(rtab3,3)
rtable(rtab4,4)
rtable(rtab5,5)
rtable(rtab6,6)
rtable(rtab7,7)


static void printRotateCol(char *src, size_t srcstep, char *dst, long lines)
{
    register char    *s, *d;
    register size_t  sstep, dstep,
                     lownyb, hinyb;
    register bit32   low, hi;

    low = hi = 0;
    
#define extract(p,t)\
    if (lines > 0) { lines--;            \
    lownyb = *p & 0x0F; hinyb = *p >> 4; \
    low |= t[lownyb]; hi |= t[hinyb]; p+= sstep;}
    
#define unpack(p,w)\
    *p = w         & 0xFF; p += dstep; \
    *p = (w >> 8)  & 0xFF; p += dstep; \
    *p = (w >> 16) & 0xFF; p += dstep; \
    *p = (w >> 24) & 0xFF; p += dstep;

    s = src; sstep = srcstep;
    d = dst; dstep = 1;

    extract(s,rtab0); extract(s,rtab1); extract(s,rtab2); extract(s,rtab3);
    extract(s,rtab4); extract(s,rtab5); extract(s,rtab6); extract(s,rtab7);
    unpack(d,hi); unpack(d,low);
}



static char       divN[256], modN[256];
static char       magic[16][16];
static Intensity *red, *green, *blue;

static int make_square(void)
{
    char   magic4x4[4][4] = { {  0, 14,  3, 13 } ,
                              { 11,  5,  8,  6 } ,
                              { 12,  2, 15,  1 } ,
                              {  7,  9,  4, 10 } };

    register int  i, j, k, l;
    double      magicfact;
    
    for (i = 0; i < 255; i++)
    {
        divN[i] = 0;
        modN[i] = i;
    }
    divN[255] = 1;
    modN[255] = 0;
    
    magicfact = 255.5 / 16.0;
    for (i = 0; i < 4; i++)
        for (j = 0; j < 4; j++)
            for (k = 0; k < 4; k++)
                for (l = 0; l < 4; l++)
                    magic[4*k+i][4*l+j] = (char)(0.5 + magic4x4[i][j] * magicfact + (magic4x4[k][l] / 16.0) * magicfact);
    return(0);
}



void orderDither8n0(char out[24])
{
    char    val;
    
    val = out[0];
    out[0] = out[1] = out[2] = out[3] = out[4] = out[5] = out[6] = out[7] = 0;
    if (val & 0x80)
        out[0] = 7;
    if (val & 0x40)
        out[1] = 7;
    if (val & 0x20)
        out[2] = 7;
    if (val & 0x10)
        out[3] = 7;
    if (val & 0x08)
        out[4] = 7;
    if (val & 0x04)
        out[5] = 7;
    if (val & 0x02)
        out[6] = 7;
    if (val & 0x01)
        out[7] = 7;
}



void orderDither8n1(char out[24], long x, long y)
{
    int     col = (int)(x & 0x0F);
    int     row = (int)(y & 0x0F);
    long    v, z, w, pixel;
    char   *in, *ex;
    
    in = out;
    ex = out;
    for (w = 0; w < 8; w++)
    {
        pixel = *in++;
        
        v = red  [pixel] >> 8;
        v = modN[v] > magic[col][row] ? divN[v] + 1 : divN[v];
        z = v;

        v = green[pixel] >> 8;
        v = modN[v] > magic[col][row] ? divN[v] + 1 : divN[v];
        v += v; z += v;

        v = blue [pixel] >> 8;
        v = modN[v] > magic[col][row] ? divN[v] + 1 : divN[v];
        v += v; v += v; z += v;

        *ex++ = ~z & 0x07;
        col++;
  }
}



void orderDither8n2(char out[24], long x, long y)
{
    int     col = (int)(x & 0x0F);
    int     row = (int)(y & 0x0F);
    long    v, z, w;
    char   *in, *ex;
    
    in = out;
    ex = out;
    for (w = 0; w < 8; w++)
    {
        v = *in++;
        v = modN[v] > magic[col][row] ? divN[v] + 1 : divN[v];
        z = v;

        v = *in++;
        v = modN[v] > magic[col][row] ? divN[v] + 1 : divN[v];
        v += v; z += v;

        v = *in++;
        v = modN[v] > magic[col][row] ? divN[v] + 1 : divN[v];
        v += v; v += v; z += v;

        *ex++ = ~z & 0x07;
        col++;
  }
}



/* 8 Zoll Papierbreite bei 300 dpi */
#define MAX_PAPERWIDTH 2400

#define INIT_GRAPH        1
#define INIT_LINE         4
#define EXIT_LINE         1
#define EXIT_GRAPH        1
#define LINE_HI           2
#define LINE_LO           1

#if INIT_GRAPH
static char init_graph[INIT_GRAPH] = { 0x10 };
#endif
#if INIT_LINE
static char init_line [INIT_LINE ] = { 0x20, 0x00, 0x00, 0x20 };
#endif
#if EXIT_LINE
static char exit_line [EXIT_LINE ] = { 0x30 };
#endif
#if EXIT_GRAPH
static char exit_graph[EXIT_GRAPH] = { 0x40 };
#endif


void gvw_main(PRINT_Structure *printS, int verbose)
{
    Image  *image;
    char   *data, out[24];
    long    width, height;
    long    output_width;
    long    linelen, planelen,
            x, y;
    int     fd, planes;
    
    image = printS->image;
    if (image == NULL)
    {
      printS->events.alert(1, "[3][ | No image to print!  | ][  Abort  ]");
      return;
    }
    data = image->data;
    if (data == NULL)
    {
      printS->events.alert(1, "[3][ | No data in image!  | ][  Abort  ]");
      return;
    }
    fd = printS->output.open("D:\\test.out");
    if (fd <= 0)
    {
      printS->events.alert(1, "[3][ | Printer not ready!  | ][  Abort  ]");
      return;
    }
    
    linelen = (long)image->width >> 3;
    output_width = image->unalignwidth;
    if (output_width > MAX_PAPERWIDTH)
      output_width = MAX_PAPERWIDTH;
    if (ATARI__TCP(image))
        linelen *= 3;
    planelen = (long)image->height * linelen;
    planes = image->depth;
#if INIT_GRAPH
    printS->output.write(fd, INIT_GRAPH, init_graph);
#endif
    make_square();
    red   = image->rgb.red;
    green = image->rgb.green;
    blue  = image->rgb.blue;

    for (y = 0, height = image->height; height > 0; height--, y++)
    {
#if INIT_LINE
        init_line[LINE_HI] = (output_width >> 8) & 0xFF;
        init_line[LINE_LO] = (output_width     ) & 0xFF;
        printS->output.write(fd, INIT_LINE, init_line);
#endif
        if (ATARIMONOP(image))
        {
            for (x = 0, width = output_width; width >= 8; width -= 8, x++) {
                out[0] = data[x];
                orderDither8n0(out);
                printS->output.write(fd, 8L, out);
            }
            if (width) {
                out[0] = data[x];
                orderDither8n0(out);
                printS->output.write(fd, width, out);
            }
        }
        else if (ATARI_RGBP(image))
        {
            for (x = 0, width = output_width; width >= 8; width -= 8, x++) {
                printRotateCol(&data[x], planelen, out, planes);
                orderDither8n1(out, x, y);
                printS->output.write(fd, 8L, out);
            }
            if (width) {
                printRotateCol(&data[x], planelen, out, planes);
                orderDither8n1(out, x, y);
                printS->output.write(fd, width, out);
            }
        }
        else if (ATARI__TCP(image))
        {
            for (x = 0, width = output_width; width >= 8; width -= 8) {
                out[ 0] = data[x++]; out[ 1] = data[x++]; out[ 2] = data[x++];
                out[ 3] = data[x++]; out[ 4] = data[x++]; out[ 5] = data[x++];
                out[ 6] = data[x++]; out[ 7] = data[x++]; out[ 8] = data[x++];
                out[ 9] = data[x++]; out[10] = data[x++]; out[11] = data[x++];
                out[12] = data[x++]; out[13] = data[x++]; out[14] = data[x++];
                out[15] = data[x++]; out[16] = data[x++]; out[17] = data[x++];
                out[18] = data[x++]; out[19] = data[x++]; out[20] = data[x++];
                out[21] = data[x++]; out[22] = data[x++]; out[23] = data[x++];
                orderDither8n2(out, x, y);
                printS->output.write(fd, 8L, out);
            }
            if (width) {
                out[ 0] = data[x++]; out[ 1] = data[x++]; out[ 2] = data[x++];
                out[ 3] = data[x++]; out[ 4] = data[x++]; out[ 5] = data[x++];
                out[ 6] = data[x++]; out[ 7] = data[x++]; out[ 8] = data[x++];
                out[ 9] = data[x++]; out[10] = data[x++]; out[11] = data[x++];
                out[12] = data[x++]; out[13] = data[x++]; out[14] = data[x++];
                out[15] = data[x++]; out[16] = data[x++]; out[17] = data[x++];
                out[18] = data[x++]; out[19] = data[x++]; out[20] = data[x++];
                out[21] = data[x++]; out[22] = data[x++]; out[23] = data[x++];
                orderDither8n2(out, x, y);
                printS->output.write(fd, width, out);
            }
        }
#if EXIT_LINE
        printS->output.write(fd, EXIT_LINE, exit_line);
#endif
        data += linelen;
    }
#if EXIT_GRAPH
    printS->output.write(fd, EXIT_GRAPH, exit_graph);
#endif
    printS->output.close(fd);
}
